<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform" xmlns:ext="http://exslt.org/common" version="1.0" extension-element-prefixes="ext">
  <xsl:output method="xml" version="1.0" encoding="UTF-8" indent="yes"/>
  <xsl:strip-space elements="*"/>
  <!-- +++ PASS 1 +++ -->
  <!-- +++ transform MuseScore 2 Measure to MuseScore 3 Measure/n*voice -->
  <!-- ‣ identity transform for most elements -->
  <xsl:template match="@* | node()" mode="pass1">
    <xsl:copy>
      <xsl:apply-templates select="@* | node()" mode="pass1"/>
    </xsl:copy>
  </xsl:template>
  <!-- ‣ split Measure element by tick element inside -->
  <xsl:template match="Measure" mode="pass1">
    <!-- copy Measure tag itself -->
    <xsl:copy>
      <!-- copy all attributes of Measure tag -->
      <xsl:apply-templates select="@*" mode="pass1"/>
      <!-- process all elements of Measure tag -->
      <xsl:call-template name="Split">
        <!-- pass tick nodes directly under Measure tag as splitters -->
        <xsl:with-param name="splitters" select="tick"/>
      </xsl:call-template>
      <!-- end of copied Measure tag -->
    </xsl:copy>
  </xsl:template>
  <!-- ‣ implementation of element multi-way split -->
  <xsl:template name="Split">
    <!-- nodes by which to split -->
    <xsl:param name="splitters"/>
    <!-- for the first run, initialise to empty node-set -->
    <xsl:param name="prefix" select="/.."/>
    <xsl:choose>
      <!-- any tick elements present? -->
      <xsl:when test="$splitters">
        <!-- cache amount of prefix nodes for later -->
        <xsl:variable name="prefixlen" select="count($prefix)"/>
        <!-- operate on first splitter (tick) node -->
        <!-- (note: this is not a loop; it sets the node list) -->
        <xsl:for-each select="$splitters[1]">
          <!-- cache all siblings preceding the splitter -->
          <xsl:variable name="nodes-upto" select="preceding-sibling::node()"/>
          <!-- create a new voice tree… -->
          <voice>
            <!-- subtract $prefix from $nodes-upto, output result -->
            <xsl:copy-of select="$nodes-upto[count(. | $prefix) != $prefixlen]"/>
          </voice>
          <xsl:choose>
            <!-- if no splitter follows… -->
            <xsl:when test="count($splitters) = 1">
              <!-- copy the rest of the Measure into the final new voice tree -->
              <voice>
                <xsl:copy-of select="current()"/>
                <xsl:copy-of select="following-sibling::node()"/>
              </voice>
            </xsl:when>
            <xsl:otherwise>
              <!-- split the rest -->
              <xsl:call-template name="Split">
                <!-- all splitters past the one we just processed -->
                <xsl:with-param name="splitters" select="$splitters[position() &gt; 1]"/>
                <!-- all nodes up to (but excluding) the current splitter -->
                <xsl:with-param name="prefix" select="$nodes-upto"/>
              </xsl:call-template>
            </xsl:otherwise>
          </xsl:choose>
        </xsl:for-each>
      </xsl:when>
      <!-- no tick elements present -->
      <xsl:otherwise>
        <!-- so just nest the Measure tag content in an extra voice tree -->
        <voice>
          <xsl:copy-of select="current()/*"/>
        </voice>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>
  <!-- +++ PASS 2 +++ -->
  <!-- +++ extract lyrics by part, voice and verse -->
  <!-- ‣ entry point for lyrics extraction -->
  <xsl:template match="/" mode="pass2">
    <lyrics>
      <!-- iterate over all parts -->
      <xsl:for-each select="//*/Score">
        <!-- skip (emit just <lyrics/>) if no lyrics (also below) -->
        <xsl:if test="count(.//Lyrics) &gt; 0">
          <!-- open tag to describe part -->
          <xsl:element name="part">
            <xsl:if test="name">
              <xsl:attribute name="name">
                <xsl:value-of select="name"/>
              </xsl:attribute>
            </xsl:if>
            <!-- iterate over all staves within that part -->
            <xsl:for-each select="Staff">
              <xsl:if test="count(.//Lyrics) &gt; 0">
                <!-- open tag to describe stave -->
                <xsl:element name="staff">
                  <xsl:attribute name="idx">
                    <xsl:value-of select="@id"/>
                  </xsl:attribute>
                  <xsl:attribute name="track">
                    <xsl:value-of select="../Part[./Staff/@id=current()/@id]/trackName"/>
                  </xsl:attribute>
                  <!-- handle lyrics of voice 1 then 2 then 3 then 4 -->
                  <xsl:call-template name="group-lyrics">
                    <xsl:with-param name="tree" select=".//Measure/voice[1]//Lyrics"/>
                    <xsl:with-param name="voice" select="1"/>
                  </xsl:call-template>
                  <xsl:call-template name="group-lyrics">
                    <xsl:with-param name="tree" select=".//Measure/voice[2]//Lyrics"/>
                    <xsl:with-param name="voice" select="2"/>
                  </xsl:call-template>
                  <xsl:call-template name="group-lyrics">
                    <xsl:with-param name="tree" select=".//Measure/voice[3]//Lyrics"/>
                    <xsl:with-param name="voice" select="3"/>
                  </xsl:call-template>
                  <xsl:call-template name="group-lyrics">
                    <xsl:with-param name="tree" select=".//Measure/voice[4]//Lyrics"/>
                    <xsl:with-param name="voice" select="4"/>
                  </xsl:call-template>
                  <!-- end of staff element -->
                </xsl:element>
              </xsl:if>
            </xsl:for-each>
            <!-- end of part element -->
          </xsl:element>
        </xsl:if>
      </xsl:for-each>
    </lyrics>
  </xsl:template>
  <!-- ‣ group lyrics nodes per verse -->
  <xsl:template name="group-lyrics">
    <!-- all nodes to operate on (e.g. Staff[1]//Measure/voice[1]//Lyrics) -->
    <xsl:param name="tree"/>
    <!-- 0-based number of current verse -->
    <xsl:param name="verse-no" select="0"/>
    <!-- current voice, for copying into verse tag attribute -->
    <xsl:param name="voice"/>
    <!-- skip output if no lyrics present -->
    <xsl:variable name="ntree" select="count($tree)"/>
    <xsl:if test="$ntree &gt; 0">
      <!-- collect all elements in the current lyrics verse: -->
      <!-- concat('0', no) prepends a 0 to the verse number if one exists… -->
      <!-- … but results in 0 if none is present (default value) -->
      <!-- number() interprets that as string describing a base-10 number -->
      <!-- so we got the actual verse number and all matching elements -->
      <xsl:variable name="elts" select="$tree[number(concat('0', no)) = $verse-no]"/>
      <xsl:variable name="nelts" select="count($elts)"/>
      <!-- open verse tag -->
      <xsl:element name="verse">
        <xsl:attribute name="voice">
          <xsl:value-of select="$voice"/>
        </xsl:attribute>
        <xsl:attribute name="nr">
          <xsl:value-of select="$verse-no + 1"/>
        </xsl:attribute>
        <!-- iterate over all Lyrics elements (see above) in that verse -->
        <xsl:for-each select="$elts">
          <!-- space if beginning of a word… but not before the first -->
          <xsl:if test="position() != 1 and not(syllabic='middle' or syllabic='end')">
            <xsl:text> </xsl:text>
          </xsl:if>
          <!-- append lyrics text to verse text node -->
          <xsl:value-of select="text"/>
          <!-- rinse and repeat -->
        </xsl:for-each>
        <!-- end of verse tag -->
      </xsl:element>
      <!-- if there are any Lyrics elements left… -->
      <xsl:if test="$ntree &gt; $nelts">
        <!-- … do that again… -->
        <xsl:call-template name="group-lyrics">
          <!-- … but remove the elements for the current verse from the list… -->
          <xsl:with-param name="tree" select="$tree[count(. | $elts) != $nelts]"/>
          <!-- … and try the next verse number (implicit sorting by verse-no) -->
          <xsl:with-param name="verse-no" select="$verse-no + 1"/>
          <xsl:with-param name="voice" select="$voice"/>
        </xsl:call-template>
      </xsl:if>
    </xsl:if>
  </xsl:template>
  <!-- +++ ENTRY POINT +++ -->
  <!-- ‣ MuseScore 2 file -->
  <xsl:template match="/museScore[starts-with(@version, '2.')]">
    <!-- run pass1 and store result node-set -->
    <xsl:variable name="rewind">
      <xsl:copy>
        <xsl:apply-templates select="@* | node()" mode="pass1"/>
      </xsl:copy>
    </xsl:variable>
    <!-- run pass2 on the node-set from that variable -->
    <xsl:apply-templates select="ext:node-set($rewind)" mode="pass2"/>
  </xsl:template>
  <!-- ‣ MuseScore 3 file -->
  <xsl:template match="/museScore[not(starts-with(@version, '2.'))]">
    <!-- just run pass2 on it -->
    <xsl:apply-templates select="/" mode="pass2"/>
  </xsl:template>
</xsl:stylesheet>
